 /*
          Jafe's Hike
          Author: HardHat (hardhat@mail.ru)

          Based on tutorial by Brad Dwyer
          Date: 2006-02-12
*/ 

#include <pspdisplay.h>
#include <pspctrl.h>
#include <pspkernel.h>
#include <pspdebug.h>
#include <pspgu.h>
#include <png.h>
#include <stdio.h>
#include <stdlib.h>
#include "graphics.h"

#define printf pspDebugScreenPrintf
#define MAX(X, Y) ((X) > (Y) ? (X) : (Y))

PSP_MODULE_INFO("Jafes Hike", 0, 1, 1);

/* Exit callback */
int exit_callback(int arg1, int arg2, void *common) {
          sceKernelExitGame();
          return 0;
}

/* Callback thread */
int CallbackThread(SceSize args, void *argp) {
          int cbid;

          cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
          sceKernelRegisterExitCallback(cbid);

          sceKernelSleepThreadCB();

          return 0;
}

/* Sets up the callback thread and returns its thread id */
int SetupCallbacks(void) {
          int thid = 0;

          thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
          if(thid >= 0) {
                    sceKernelStartThread(thid, 0, 0);
          }

          return thid;
}

Image *mirrorImage(Image *in)
{
	if(!in) return 0;
	Image *out=createImage(in->imageWidth,in->imageHeight);
	if(!out) return 0;
	Color *a,*b;
	int x,y;

	a=in->data;
	b=out->data;
	for(y=0;y<in->imageHeight;y++) {
		for(x=0;x<in->imageWidth;x++) {
			*b=a[in->imageWidth-x-1];
			b++;
		}
		a+=in->textureWidth;
	}
	return out;
}

struct Butterfly {
	int x,y;
	int dx,dy;
	int delay;
	int frame;
	int attraction;
} bfly[65];
int activeBfly;
Image *bflyImage[6];
int score;
int level;

void grabBfly(int x,int y)
{
	int i;
	
	for(i=0;i<activeBfly;i++) {
		// must be in range: { x+32 < bx+32 && bx < x+128-32 }
		// { y < by+32 , by < y+32 }
		if(bfly[i].x+32<x+32 || bfly[i].x>x+128-32 || bfly[i].y+32<y || bfly[i].y>y+32) 
			continue;
		if(bfly[i].frame>3) continue;	// Already disappearing
		// We got one!  Time to sparkle.
		score++;
#if 0
		break;
#else
		bfly[i].frame=4;	// Sparkle time.
		bfly[i].delay=40;
		bfly[i].dx=0;
		bfly[i].dy=2;
		break;			// Only one per grab call.
#endif
	}
}

void newBfly(int x,int y)
{
	if(activeBfly>=64) return;

	bfly[activeBfly].y=-16;
	bfly[activeBfly].x=x>240?-32:480;
	bfly[activeBfly].dx=0;		// Will choose direction later.
	bfly[activeBfly].dy=0;
	bfly[activeBfly].delay=1;
	bfly[activeBfly].frame=0;
	bfly[activeBfly].attraction=128;	// mixed

	activeBfly++;
}

/* quickrand, copied from Pang_for_PSP source by Michel Jouvet. */

static unsigned Seed1;
static unsigned Seed2;
unsigned quickrand(unsigned long MAX)
{
        Seed1 = (Seed1 + 46906481) ^ Seed2;
        Seed2 = Seed1 ^ ( ((Seed2<<3) | (Seed2 >> 29)) + 103995407);

        return (Seed1 - Seed2)%MAX;
}

void headingBfly(int x,int y,int i)
{
	// Attraction drops with proximity to Jeffrey.
	//int r=rand()&255;
	int r=quickrand(256);
	int a=bfly[i].attraction;
	if(bfly[i].y<y)  {
		if(r<a) bfly[i].dy=1;
		else if(r<128) bfly[i].dy=-1;
		else bfly[i].dy=0;
	} else if(bfly[i].y>y+32) {
		if(r<a) bfly[i].dy=-1;
		else if(r<128) bfly[i].dy=1;
		else bfly[i].dy=0;
	}
	if(bfly[i].x<x)  {
		if(r<a) bfly[i].dx=1;
		else if(r<128) bfly[i].dx=-1;
		else bfly[i].dx=0;
	} else if(bfly[i].y>y+32) {
		if(r<a) bfly[i].dx=-1;
		else if(r<128) bfly[i].dx=1;
		else bfly[i].dx=0;
	}
	// Now calculate distance
	int dist;
	dist=bfly[i].y<y?y-bfly[i].y:bfly[i].y-y;
	dist+=bfly[i].x<x+64?x+64-bfly[i].x:bfly[i].x-64-x;
	if(dist<bfly[i].attraction) {
		bfly[i].attraction=dist;
	}
	if(dist<64) bfly[i].attraction=64;	// Balance nearby
	if(dist>255) bfly[i].attraction=255;	// Balance far
}

void nextBfly(int x,int y)
{
	int i;
	
	while(activeBfly<level && activeBfly<64) {
		newBfly(x,y);
	}
	for(i=0;i<activeBfly;i++) {
		if(bfly[i].delay<1) {
			if(bfly[i].frame<4) {
				bfly[i].delay=5;
				bfly[i].frame=(bfly[i].frame+1)%2;
				// Now choose the new direction.
				headingBfly(x,y,i);
				if(bfly[i].dx<0) {
					bfly[i].frame|=2;
				}
			} else {
				// Captured butterfly is about to disappear
				bfly[i].frame=6;	// Mark for deletion
			}
		} else {
			bfly[i].delay--;
			if(bfly[i].frame>=4 && bfly[i].delay&1) {
				bfly[i].frame^=1;
			}
		}
		bfly[i].x+=bfly[i].dx;
		bfly[i].y+=bfly[i].dy;
	}
	for(i=0;i<activeBfly;i++) {
		if(bfly[i].frame>=6) {
			// Dead.
			int j;
			for(j=i+1;j<activeBfly;j++) {
				bfly[j-1]=bfly[j];
			}
			activeBfly--;
		}
	}
}

void drawBfly()
{
	int i;
	
	for(i=0;i<activeBfly;i++) {
		if(bfly[i].frame<6 && bflyImage[bfly[i].frame]) {
			int wid=bflyImage[bfly[i].frame]->imageWidth;
			int hei=bflyImage[bfly[i].frame]->imageHeight;

			blitAlphaImageToScreen(0,0,wid,hei,bflyImage[bfly[i].frame],bfly[i].x,bfly[i].y);
		} else {
			static Color choice=0xff00d0ff;
			fillScreenRect(choice,bfly[i].x,bfly[i].y,32,32);
		}
	}
}

/// Note:  only works in the special case of a horizontal or vertical line
void clipLineScreen(int x0,int y0,int x1,int y1,Color color)
{
	if(x0==x1) {
		if(x0<0 || x1>=SCREEN_WIDTH) return;
	} else {
		if(x0<0 && x1<0) return;
		if(x0>=SCREEN_WIDTH && x1>=SCREEN_WIDTH) return;
		// What is left is that some of the line is onscreen in the x axis.
		if(x0<0) x0=0;
		if(x0>=SCREEN_WIDTH) x0=SCREEN_WIDTH-1;
		if(x1<0) x1=0;
		if(x1>=SCREEN_WIDTH) x1=SCREEN_WIDTH-1;
	}
	if(y0==y1) {
		if(y0<0 || y1>=SCREEN_HEIGHT) return;
	} else {
		if(y0<0 && y1<0) return;
		if(y0>=SCREEN_HEIGHT && y1>=SCREEN_HEIGHT) return;
		// What is left is that some of the line is onscreen in the y axis.
		if(y0<0) y0=0;
		if(y0>=SCREEN_HEIGHT) y0=SCREEN_HEIGHT-1;
		if(y1<0) y1=0;
		if(y1>=SCREEN_HEIGHT) y1=SCREEN_HEIGHT-1;
	}
	// We made it here, so there is at least one pixel to show.
	drawLineScreen(x0,y0,x1,y1,color);
}

int main() 
{
	Image* bgImage=0,*poseImage[6];
	int jeffx,jeffy,jeffframe;
	int jump,energy;
	int delay;
	int targetRect=0;
	int oldButtons=0;
	SceCtrlData pad;
	char *message=0;
	int messageTimer=0;
	
	pspDebugScreenInit();
	SetupCallbacks();
	printf("     Jafe's Hike\n\nBy HardHat\n\n\nUse [X] to hop\n"
	"Use [[]] to catch butterflies\n"
	"Toggle the targeting rectangle with [/\\]\n"
	"Screen capture with SELECT\n"
	"Pause with HOLD\n"
	"Arrows to move left and right\n"
	"Up and down controls the number of butterflies\n");
	printf("\nPress [X] to start.");
	do {
		sceDisplayWaitVblankStart();
		sceCtrlReadBufferPositive(&pad,1);
	} while(!(pad.Buttons & PSP_CTRL_CROSS));
	printf("\nRelease [X] to start.");
	while(pad.Buttons & PSP_CTRL_CROSS) {
		sceDisplayWaitVblankStart();
		sceCtrlReadBufferPositive(&pad,1);
	} 
	printf("\nInitializing graphics...\n");
	initGraphics();
	
	bgImage = loadImage("bg.png");
	poseImage[0] = loadImage("walk1.png");
	poseImage[1] = loadImage("walk2.png");
	poseImage[2] = mirrorImage(poseImage[0]);
	poseImage[3] = mirrorImage(poseImage[1]);
	poseImage[4] = loadImage("standing.png");
	poseImage[5] = loadImage("catch.png");
	bflyImage[0] = loadImage("bfly1.png");
	bflyImage[1] = loadImage("bfly2.png");
	bflyImage[2] = mirrorImage(bflyImage[0]);
	bflyImage[3] = mirrorImage(bflyImage[1]);
	bflyImage[4] = loadImage("sparkle1.png");
	bflyImage[5] = loadImage("sparkle2.png");

	//srand(33456);

	jeffx=240;
	jeffy=16;
	jeffframe=4;
	delay=1;
	level=3;
	jump=0;
	energy=0;
	while(1) {
		sceDisplayWaitVblankStart();
		sceCtrlReadBufferPositive(&pad,1);
		if(bgImage) {
			blitAlphaImageToScreen(0 ,0 ,480 , 272, bgImage, 0, 0);
		} else {
			fillScreenRect(0xffff8000,0,0,480,272);
		}


		if(pad.Buttons & PSP_CTRL_LEFT) {
			if(delay--<=0) {
				jeffframe=((jeffframe+1)%2)+2;
				delay=8;
			}
			jeffx--;
			if(jeffx==-128) jeffx=480;
		} else if(pad.Buttons & PSP_CTRL_RIGHT) {
			if(delay--<=0) {
				jeffframe=(jeffframe+1)%2;
				delay=8;
			}
			jeffx++;
			if(jeffx==480) jeffx=-128;
		} else if(pad.Buttons & PSP_CTRL_SQUARE) {
			jeffframe=5;
			grabBfly(jeffx,jeffy);
		} else {
			jeffframe=4;
		}
		// Debug only:
		static int levelChanged=0;
		if(pad.Buttons & PSP_CTRL_UP) {
			if(!levelChanged) level++;
			levelChanged=1;
		} else if(pad.Buttons & PSP_CTRL_DOWN) {
			if((!levelChanged) && level>1)
				level--;
			levelChanged=1;
		} else {
			levelChanged=0;
		}
		if(pad.Buttons &PSP_CTRL_TRIANGLE && !(oldButtons&PSP_CTRL_TRIANGLE) ) {
			// Turn on targeting.
			targetRect^=1;
		}
		if(pad.Buttons &PSP_CTRL_SELECT && !(oldButtons&PSP_CTRL_SELECT) ) {
			// Save a screenshot.
			static int id=0;
			id++;
			static char buf[128];
			sprintf(buf,"jafehike%03d.png",id);
			saveImage(buf,getVramDisplayBuffer(),SCREEN_WIDTH,SCREEN_HEIGHT,PSP_LINE_SIZE,0);
			sprintf(buf,"Saved screen shot: jafehike%03d.png",id);
			message=buf;
			messageTimer=360;
		}
		if((!jump) && poseImage[0] && (pad.Buttons & PSP_CTRL_SQUARE)) {
			jump=1;
			energy=10;
		} else if((!jump) && (pad.Buttons & PSP_CTRL_CROSS)) {
			jump=1;
			energy=5;
		} else if(jump==1 && energy<0) {
			jump=0;
		} else if(jump) {
			jeffy-=energy;
			jump+=energy;
			energy--;
		}
		
		oldButtons=pad.Buttons;

		nextBfly(jeffx,jeffy);
		if(poseImage[jeffframe]) {
			int wid=poseImage[jeffframe]->imageWidth;
			int hei=poseImage[jeffframe]->imageHeight;

			blitAlphaImageToScreen(0,0,wid,hei,poseImage[jeffframe],jeffx,jeffy);
		} else {
			static Color choice[]={0xffff7f00,0xff7f7f00,0xffffffff,0xff000000,0xff80ff00,0xffff00ff};
			fillScreenRect(choice[jeffframe],jeffx,jeffy,128,256);
		}
		drawBfly();

		// Draw a targeting rect if requested.
		if(targetRect) {
			clipLineScreen(jeffx+32,jeffy,jeffx+128-32,jeffy,0xffff00ff);
			clipLineScreen(jeffx+128-32,jeffy,jeffx+128-32,jeffy+32,0xffff00ff);
			clipLineScreen(jeffx+32,jeffy+32,jeffx+128-32,jeffy+32,0xffff00ff);
			clipLineScreen(jeffx+32,jeffy,jeffx+32,jeffy+32,0xffff00ff);
		}
		char buffer[256];
		sprintf(buffer,"Level %d Score: %d.  Active butterflies: %d",level,score,activeBfly);
		fillScreenRect(24,239,480-24,249,0xffc0c0c0);
		printTextScreen(26,241,buffer,0xff000000);
		printTextScreen(25,240,buffer,0xff0000ff);
		if(pad.Buttons & PSP_CTRL_HOLD) {
			message="Game paused.";
			messageTimer=1;
		}
		if(messageTimer>0) {
			messageTimer--;
			fillScreenRect(24,229,480-24,239,0xffc0c0c0);
			printTextScreen(26,231,message,0xff000000);
			printTextScreen(25,230,message,0xff0000ff);
		}
		flipScreen();
		while(pad.Buttons & PSP_CTRL_HOLD) {
			sceDisplayWaitVblankStart();
			sceCtrlReadBufferPositive(&pad,1);
		}
	}

	sceKernelSleepThread();
	return 0;
}

